@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_exec -- a module for executing external scripts
  *
- * Copyright (c) 2002-2010 TJ Saunders
+ * Copyright (c) 2002-2011 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders gives permission to link this program
  * with OpenSSL, and distribute the resulting executable, without including
@@ -24,20 +24,21 @@
  * This is mod_exec, contrib software for proftpd 1.3.x and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_exec.c,v 1.9.2.1 2010/04/12 17:09:29 castaglia Exp $
+ * $Id: mod_exec.c,v 1.20 2011/09/24 06:44:36 castaglia Exp $
  */
 
 #include "conf.h"
 #include "privs.h"
 
-#include <signal.h>
-#include <sys/resource.h>
+#ifdef HAVE_SYS_RESOURCE_H
+# include <sys/resource.h>
+#endif
 
-#define MOD_EXEC_VERSION	"mod_exec/0.9.9"
+#define MOD_EXEC_VERSION	"mod_exec/0.9.11"
 
 /* Make sure the version of proftpd is as necessary. */
-#if PROFTPD_VERSION_NUMBER < 0x0001030301
-# error "ProFTPD 1.3.3rc1 or later required"
+#if PROFTPD_VERSION_NUMBER < 0x0001030402
+# error "ProFTPD 1.3.4rc2 or later required"
 #endif
 
 module exec_module;
@@ -168,7 +169,7 @@ static unsigned char exec_match_cmd(cmd_rec *cmd, array_header *cmd_array) {
         strcasecmp(cmds[i], cmd->group) == 0)
       return TRUE;
 
-    if (strcasecmp(cmds[i], "ALL") == 0)
+    if (strncasecmp(cmds[i], "ALL", 4) == 0)
       return TRUE;
   }
 
@@ -184,7 +185,7 @@ static int exec_openlog(void) {
     return 0;
 
   /* Check for "none". */
-  if (strcasecmp(exec_logname, "none") == 0) {
+  if (strncasecmp(exec_logname, "none", 5) == 0) {
     exec_logname = NULL;
     return 0;
   }
@@ -226,7 +227,7 @@ static char **exec_prepare_environ(pool *env_pool, cmd_rec *cmd) {
   while (c) {
     pr_signals_handle();
 
-    if (strcmp("-", c->argv[1]) == 0) {
+    if (strncmp("-", c->argv[1], 2) == 0) {
       *((char **) push_array(env)) = pstrcat(env_pool, c->argv[0], "=",
         getenv(c->argv[0]) ? getenv(c->argv[0]) : "", NULL);
 
@@ -597,7 +598,7 @@ static int exec_ssystem(cmd_rec *cmd, config_rec *c, int flags) {
             (char *) c->argv[i], exec_stdin_pipe[1]);
         }
 
-        if (write(exec_stdin_pipe[1], "\n", 2) < 0) {
+        if (write(exec_stdin_pipe[1], "\n", 1) < 0) {
           exec_log("error writing newline to stdin: %s", strerror(errno));
         }
       }
@@ -616,14 +617,14 @@ static int exec_ssystem(cmd_rec *cmd, config_rec *c, int flags) {
     }
 
     if (exec_opts & EXEC_OPT_USE_STDIN) {
-      close(exec_stdin_pipe[0]);
+      (void) close(exec_stdin_pipe[0]);
       exec_stdin_pipe[0] = -1;
     }
 
-    close(exec_stdout_pipe[1]);
+    (void) close(exec_stdout_pipe[1]);
     exec_stdout_pipe[1] = -1;
 
-    close(exec_stderr_pipe[1]);
+    (void) close(exec_stderr_pipe[1]);
     exec_stderr_pipe[1] = -1;
    
     if ((exec_opts & EXEC_OPT_LOG_STDOUT) ||
@@ -836,12 +837,23 @@ static int exec_ssystem(cmd_rec *cmd, config_rec *c, int flags) {
     return errno;
   }
 
-  if (WIFEXITED(status))
-    return WEXITSTATUS(status);
+  if (WIFEXITED(status)) {
+    int exit_status;
+
+    exit_status = WEXITSTATUS(status);
+    exec_log("'%s' terminated normally, with exit status %d",
+      (const char *) c->argv[2], exit_status);
+    return exit_status;
+  }
 
   if (WIFSIGNALED(status)) {
     exec_log("'%s' died from signal %d", (const char *) c->argv[2],
       WTERMSIG(status));
+
+    if (WCOREDUMP(status)) {
+      exec_log("'%s' created a coredump", (const char *) c->argv[2]);
+    }
+
     return EPERM;
   }
 
@@ -863,6 +875,18 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
         pr_netaddr_get_ipstr(remote_addr) : "", NULL);
   }
 
+  ptr = strstr(varstr, "%A");
+  if (ptr != NULL) {
+    char *anon_pass;
+
+    anon_pass = pr_table_get(session.notes, "mod_auth.anon-passwd", NULL);
+    if (anon_pass == NULL) {
+      anon_pass = "UNKNOWN";
+    }
+
+    varstr = sreplace(tmp_pool, varstr, "%A", anon_pass, NULL);
+  }
+
   ptr = strstr(varstr, "%C");
   if (ptr != NULL) {
     varstr = sreplace(tmp_pool, varstr, "%C",
@@ -877,7 +901,7 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
 
   ptr = strstr(varstr, "%F");
   if (ptr != NULL) {
-    if (strcmp(cmd->argv[0], C_RNTO) == 0) {
+    if (pr_cmd_cmp(cmd, PR_CMD_RNTO_ID) == 0) {
       char *path;
 
       path = dir_best_path(tmp_pool, pr_fs_decode_path(tmp_pool, cmd->arg));
@@ -888,9 +912,9 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
       varstr = sreplace(tmp_pool, varstr, "%F", session.xfer.path, NULL);
 
     } else if (session.curr_phase == PRE_CMD &&
-               (strcmp(cmd->argv[0], C_STOR) == 0 ||
-                strcmp(cmd->argv[0], C_RETR) == 0 ||
-                strcmp(cmd->argv[0], C_APPE) == 0)) {
+               (pr_cmd_cmp(cmd, PR_CMD_STOR_ID) == 0 ||
+                pr_cmd_cmp(cmd, PR_CMD_RETR_ID) == 0 ||
+                pr_cmd_cmp(cmd, PR_CMD_APPE_ID) == 0)) {
       char *path;
 
       /* If we're in the PRE_CMD phase, then the %f variable can't be
@@ -904,7 +928,7 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
        * stored in the session.xfer structure; these should be expanded
        * properly as well.
        */
-      if (strcmp(cmd->argv[0], C_DELE) == 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_DELE_ID) == 0) {
         char *path;
 
         path = dir_best_path(tmp_pool, pr_fs_decode_path(tmp_pool, cmd->arg));
@@ -919,7 +943,7 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
   ptr = strstr(varstr, "%f");
   if (ptr != NULL) {
 
-    if (strcmp(cmd->argv[0], C_RNTO) == 0) {
+    if (pr_cmd_cmp(cmd, PR_CMD_RNTO_ID) == 0) {
       char *path;
 
       path = pr_fs_decode_path(tmp_pool, cmd->arg);
@@ -932,9 +956,9 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
         dir_abs_path(tmp_pool, session.xfer.path, TRUE), NULL);
 
     } else if (session.curr_phase == PRE_CMD &&
-               (strcmp(cmd->argv[0], C_STOR) == 0 ||
-                strcmp(cmd->argv[0], C_RETR) == 0 ||
-                strcmp(cmd->argv[0], C_APPE) == 0)) {
+               (pr_cmd_cmp(cmd, PR_CMD_STOR_ID) == 0 ||
+                pr_cmd_cmp(cmd, PR_CMD_RETR_ID) == 0 ||
+                pr_cmd_cmp(cmd, PR_CMD_APPE_ID) == 0)) {
       char *path;
 
       /* If we're in the PRE_CMD phase, then the %f variable can't be
@@ -950,11 +974,11 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
        * filenames that are not stored in the session.xfer structure; these
        * should be expanded properly as well.
        */
-      if (strcmp(cmd->argv[0], C_DELE) == 0 ||
-          strcmp(cmd->argv[0], C_MKD) == 0 ||
-          strcmp(cmd->argv[0], C_RMD) == 0 ||
-          strcmp(cmd->argv[0], C_XMKD) == 0 ||
-          strcmp(cmd->argv[0], C_XRMD) == 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_DELE_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_MKD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_RMD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XMKD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XRMD_ID) == 0) {
         varstr = sreplace(tmp_pool, varstr, "%f",
           dir_abs_path(tmp_pool, cmd->arg, TRUE), NULL);
 
@@ -1000,7 +1024,7 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
   ptr = strstr(varstr, "%r");
   if (ptr != NULL) {
     if (cmd) {
-      if (strcasecmp(cmd->argv[0], C_PASS) == 0 &&
+      if (pr_cmd_cmp(cmd, PR_CMD_PASS_ID) == 0 &&
           session.hide_password) {
         varstr = sreplace(tmp_pool, varstr, "%r", "PASS (hidden)", NULL);
 
@@ -1042,7 +1066,7 @@ static char *exec_subst_var(pool *tmp_pool, char *varstr, cmd_rec *cmd) {
   if (ptr != NULL) {
     char *rnfr_path = "-";
 
-    if (strcmp(cmd->argv[0], C_RNTO) == 0) {
+    if (pr_cmd_cmp(cmd, PR_CMD_RNTO_ID) == 0) {
       rnfr_path = pr_table_get(session.notes, "mod_core.rnfr-path", NULL);
       if (rnfr_path == NULL)
         rnfr_path = "-";
@@ -1371,7 +1395,6 @@ MODRET set_execenviron(cmd_rec *cmd) {
   for (i = 0; i < strlen(cmd->argv[1]); i++)
     (cmd->argv[1])[i] = toupper((cmd->argv[1])[i]);
   c->argv[0] = pstrdup(c->pool, cmd->argv[1]);
-  c->flags |= CF_MERGEDOWN_MULTI;
 
   return PR_HANDLED(cmd);
 }
@@ -1514,11 +1537,11 @@ MODRET set_execonevent(cmd_rec *cmd) {
   eed->event = pstrdup(c->pool, cmd->argv[1]);
   eed->c = c;
 
-  if (strcasecmp(eed->event, "MaxConnectionRate") == 0) {
+  if (strncasecmp(eed->event, "MaxConnectionRate", 18) == 0) {
     pr_event_register(&exec_module, "core.max-connection-rate", exec_any_ev,
       eed);
 
-  } else if (strcasecmp(eed->event, "MaxInstances") == 0) {
+  } else if (strncasecmp(eed->event, "MaxInstances", 13) == 0) {
      pr_event_register(&exec_module, "core.max-instances", exec_any_ev, eed);
 
   } else
@@ -1598,16 +1621,16 @@ MODRET set_execoptions(cmd_rec *cmd) {
   c = add_config_param(cmd->argv[0], 1, NULL);
 
   for (i = 1; i < cmd->argc; i++) {
-    if (strcmp(cmd->argv[i], "logStdout") == 0) {
+    if (strncmp(cmd->argv[i], "logStdout", 10) == 0) {
       opts |= EXEC_OPT_LOG_STDOUT;
 
-    } else if (strcmp(cmd->argv[i], "logStderr") == 0) {
+    } else if (strncmp(cmd->argv[i], "logStderr", 10) == 0) {
       opts |= EXEC_OPT_LOG_STDERR;
 
-    } else if (strcmp(cmd->argv[i], "sendStdout") == 0) {
+    } else if (strncmp(cmd->argv[i], "sendStdout", 11) == 0) {
       opts |= EXEC_OPT_SEND_STDOUT;
 
-    } else if (strcmp(cmd->argv[i], "useStdin") == 0) {
+    } else if (strncmp(cmd->argv[i], "useStdin", 9) == 0) {
       opts |= EXEC_OPT_USE_STDIN;
 
     } else {
@@ -1694,7 +1717,7 @@ static void exec_exit_ev(const void *event_data, void *user_data) {
 
 #if defined(PR_SHARED_MODULE)
 static void exec_mod_unload_ev(const void *event_data, void *user_data) {
-  if (strcmp("mod_exec.c", (const char *) event_data) == 0) {
+  if (strncmp("mod_exec.c", (const char *) event_data, 11) == 0) {
     if (exec_pool) {
       destroy_pool(exec_pool);
       exec_pool = NULL;
@@ -1772,9 +1795,11 @@ static void exec_restart_ev(const void *event_data, void *user_data) {
 static int exec_sess_init(void) {
   unsigned char *use_exec = NULL;
   config_rec *c = NULL;
+  const char *proto;
 
- if ((use_exec = get_param_ptr(main_server->conf, "ExecEngine",
-    FALSE)) != NULL && *use_exec == TRUE) {
+  use_exec = get_param_ptr(main_server->conf, "ExecEngine", FALSE);
+  if (use_exec != NULL &&
+      *use_exec == TRUE) {
     exec_engine = TRUE;
 
   } else {
@@ -1782,11 +1807,25 @@ static int exec_sess_init(void) {
     return 0;
   }
 
+  /* Register a "core.exit" event handler. */
+  pr_event_register(&exec_module, "core.exit", exec_exit_ev, NULL);
+
   c = find_config(main_server->conf, CONF_PARAM, "ExecOptions", FALSE);
   if (c) {
     exec_opts = *((unsigned int *) c->argv[0]);
   }
 
+  /* If we are handling an SSH2 session, then disable the sendStdout
+   * ExecOption, if present.
+   *
+   * Attempting to send the stdout of commands to connecting SSH2 clients
+   * can confuse them and lead to connection problems.
+   */
+  proto = pr_session_get_protocol(0);
+  if (strncmp(proto, "ssh2", 5) == 0) {
+    exec_opts &= ~EXEC_OPT_SEND_STDOUT;
+  }
+
   c = find_config(main_server->conf, CONF_PARAM, "ExecTimeout", FALSE);
   if (c)
     exec_timeout = *((time_t *) c->argv[0]);
@@ -1824,9 +1863,6 @@ static int exec_sess_init(void) {
     c = find_config_next(c, c->next, CONF_PARAM, "ExecOnConnect", FALSE);
   }
 
-  /* Register a "core.exit" event handler. */
-  pr_event_register(&exec_module, "core.exit", exec_exit_ev, NULL);
-
   return 0;
 }
 
