@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2003-2006 The ProFTPD Project team
+ * Copyright (c) 2003-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,7 +23,7 @@
  */
 
 /* Network address routines
- * $Id: netaddr.c,v 1.58 2006/12/19 21:36:30 castaglia Exp $
+ * $Id: netaddr.c,v 1.79 2011/05/23 21:22:24 castaglia Exp $
  */
 
 #include "conf.h"
@@ -31,7 +31,7 @@
 /* Define an IPv4 equivalent of the IN6_IS_ADDR_LOOPBACK macro. */
 #undef IN_IS_ADDR_LOOPBACK
 #define IN_IS_ADDR_LOOPBACK(a) \
-  ((((long int) (a)->s_addr) & 0xff000000) == 0x7f000000)
+  ((((unsigned long int) ntohl((a)->s_addr)) & 0xff000000) == 0x7f000000)
 
 static pr_netaddr_t sess_local_addr;
 static int have_sess_local_addr = FALSE;
@@ -50,8 +50,163 @@ static int use_ipv6 = TRUE;
 static int use_ipv6 = FALSE;
 #endif /* PR_USE_IPV6 */
 
+static char localaddr_str[PR_TUNABLE_BUFFER_SIZE];
+static int have_localaddr_str = FALSE;
+
+static pool *netaddr_pool = NULL;
+static pr_table_t *netaddr_iptab = NULL;
+static pr_table_t *netaddr_dnstab = NULL;
+
 static const char *trace_channel = "dns";
 
+/* Netaddr cache management */
+static array_header *netaddr_dnscache_get(pool *p, const char *ip_str) {
+  array_header *res = NULL;
+
+  if (netaddr_dnstab) {
+    void *v = pr_table_get(netaddr_dnstab, ip_str, NULL);
+    if (v) {
+      res = v;
+
+      pr_trace_msg(trace_channel, 4,
+        "using %d DNS %s from netaddr DNS cache for IP address '%s'",
+        res->nelts, res->nelts != 1 ? "names" : "name", ip_str);
+
+      if (p) {
+        /* If the caller provided a pool, return a copy of the array. */
+        return copy_array_str(p, res);
+      }
+
+      return res;
+    }
+  }
+
+  pr_trace_msg(trace_channel, 12,
+    "no DNS names found in netaddr DNS cache for IP address '%s'", ip_str);
+  errno = ENOENT;
+  return NULL;
+}
+
+static void netaddr_dnscache_set(const char *ip_str, const char *dns_name) {
+  if (netaddr_dnstab) {
+    void *v = NULL;
+    array_header *res = NULL;
+    int add_list = FALSE;
+
+    res = netaddr_dnscache_get(NULL, ip_str);
+    if (res == NULL) {
+      /* No existing entries for this IP address yet. */
+      res = make_array(netaddr_pool, 1, sizeof(char *));
+      add_list = TRUE;
+
+    } else {
+      register unsigned int i;
+      char **names;
+
+      /* Check for duplicates. */
+      names = res->elts;
+      for (i = 0; i < res->nelts; i++) {
+        if (names[i] != NULL) {
+          if (strcmp(names[i], dns_name) == 0) {
+            pr_trace_msg(trace_channel, 5,
+              "DNS name '%s' for IP address '%s' already stashed in the "
+              "netaddr DNS cache", dns_name, ip_str);
+            return;
+          }
+        }
+      }
+    }
+
+    *((char **) push_array(res)) = pstrdup(netaddr_pool, dns_name);
+    v = res;
+
+    if (add_list) { 
+      if (pr_table_add(netaddr_dnstab, pstrdup(netaddr_pool, ip_str), v,
+          sizeof(array_header *)) < 0) {
+        pr_trace_msg(trace_channel, 3,
+          "error adding DNS name '%s' for IP address '%s' to the netaddr "
+          "DNS cache: %s", dns_name, ip_str, strerror(errno));
+
+      } else {
+        pr_trace_msg(trace_channel, 5,
+          "stashed DNS name '%s' for IP address '%s' in the netaddr DNS cache",
+          dns_name, ip_str);
+      }
+
+    } else {
+      pr_trace_msg(trace_channel, 5,
+        "stashed DNS name '%s' for IP address '%s' in the netaddr DNS cache",
+        dns_name, ip_str);
+    }
+  }
+
+  return;
+}
+
+static pr_netaddr_t *netaddr_ipcache_get(pool *p, const char *name) {
+  pr_netaddr_t *res = NULL;
+
+  if (netaddr_iptab) {
+    void *v = pr_table_get(netaddr_iptab, name, NULL);
+    if (v) {
+      res = v;
+      pr_trace_msg(trace_channel, 4,
+        "using IP address '%s' from netaddr IP cache for name '%s'",
+        pr_netaddr_get_ipstr(res), name);
+
+      /* We return a copy of the cache's netaddr_t, if the caller provided
+       * a pool for duplication.
+       */
+      if (p) {
+        return pr_netaddr_dup(p, res);
+      }
+
+      return res;
+    }
+  }
+
+  pr_trace_msg(trace_channel, 2,
+    "no IP address found in netaddr IP cache for name '%s'", name);
+  errno = ENOENT;
+  return NULL;
+}
+
+static void netaddr_ipcache_set(const char *name, pr_netaddr_t *na) {
+  if (netaddr_iptab) {
+    int count = 0;
+    void *v = NULL;
+
+    /* We store an internal copy of the netaddr_t in the cache. */
+    v = pr_netaddr_dup(netaddr_pool, na);
+
+    count = pr_table_exists(netaddr_iptab, name);
+    if (count <= 0) {
+      if (pr_table_add(netaddr_iptab, pstrdup(netaddr_pool, name), v,
+          sizeof(pr_netaddr_t *)) < 0) {
+        pr_trace_msg(trace_channel, 3,
+          "error adding IP address '%s' for name '%s' to the netaddr "
+          "IP cache: %s", pr_netaddr_get_ipstr(na), name,
+          strerror(errno));
+
+      } else {
+        pr_trace_msg(trace_channel, 5,
+          "stashed IP address '%s' for name '%s' in the netaddr IP cache",
+          pr_netaddr_get_ipstr(v), name);
+      }
+
+    } else {
+      if (pr_table_set(netaddr_iptab, pstrdup(netaddr_pool, name), v,
+          sizeof(pr_netaddr_t *)) < 0) {
+        pr_trace_msg(trace_channel, 3,
+          "error setting IP address '%s' for name '%s' in the netaddr "
+          "IP cache: %s", pr_netaddr_get_ipstr(na), name, strerror(errno));
+      }
+    }
+  }
+
+  return;
+}
+
 /* Provide replacements for needed functions. */
 
 #if !defined(HAVE_GETNAMEINFO) || defined(PR_USE_GETNAMEINFO)
@@ -137,11 +292,12 @@ int pr_getaddrinfo(const char *node, const char *service,
          *pe->p_name != 0)
       proto_name = pe->p_name;
 
-    if (hints->ai_socktype != 0)
+    if (hints->ai_socktype != 0) {
       socktype = hints->ai_socktype;
 
-    else if (strcasecmp(proto_name, "udp") == 0)
+    } else if (strncasecmp(proto_name, "udp", 4) == 0) {
       socktype = SOCK_DGRAM;
+    }
   }
 
   if (service != NULL) {
@@ -243,6 +399,39 @@ static void *get_v4inaddr(const pr_netaddr_t *na) {
   return (((char *) pr_netaddr_get_inaddr(na)) + 12);
 }
 
+/* Validate anything returned from the 'outside', since it's untrusted
+ * information.
+ */
+char *pr_netaddr_validate_dns_str(char *buf) {
+  char *p;
+
+  if (buf == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  /* Validate anything returned from a DNS. */
+  for (p = buf; p && *p; p++) {
+
+    /* Per RFC requirements, these are all that are valid from a DNS. */
+    if (!isalnum((int) *p) &&
+        *p != '.' &&
+        *p != '-'
+#ifdef PR_USE_IPV6
+        && *p != ':'
+#endif /* PR_USE_IPV6 */
+        ) {
+
+      /* We set it to _ because we know that's an invalid, yet safe, option
+       * for a DNS entry.
+       */
+      *p = '_';
+    }
+  }
+
+  return buf;
+}
+
 int pr_netaddr_set_reverse_dns(int enable) {
   int old_enable = reverse_dns;
   reverse_dns = enable;
@@ -278,6 +467,16 @@ pr_netaddr_t *pr_netaddr_dup(pool *p, pr_netaddr_t *na) {
   pr_netaddr_set_family(dup_na, pr_netaddr_get_family(na));
   pr_netaddr_set_sockaddr(dup_na, pr_netaddr_get_sockaddr(na));  
 
+  if (na->na_have_ipstr) {
+    sstrncpy(dup_na->na_ipstr, na->na_ipstr, sizeof(dup_na->na_ipstr));
+    dup_na->na_have_ipstr = 1;
+  }
+
+  if (na->na_have_dnsstr) {
+    sstrncpy(dup_na->na_dnsstr, na->na_dnsstr, sizeof(dup_na->na_dnsstr));
+    dup_na->na_have_dnsstr = 1;
+  }
+
   return dup_na;
 }
 
@@ -288,7 +487,8 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
   pr_netaddr_t *na = NULL;
   int res;
 
-  if (p == NULL || name == NULL) {
+  if (p == NULL ||
+      name == NULL) {
     errno = EINVAL;
     return NULL;
   }
@@ -296,6 +496,20 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
   pr_trace_msg(trace_channel, 10, "resolving name '%s' to IP address",
     name);
 
+  /* First, check our cache to see if this name has already been
+   * resolved.  We only want to use the cache, though, if the caller did not
+   * provide the `addrs' pointer, indicating that the caller wants to know
+   * about any additional addresses for the given name.  The netaddr cache
+   * is a simple cache, hidden from callers, and thus is unable to populate
+   * that `addrs' pointer if the name is in the cache.
+   */
+  if (addrs == NULL) {
+    na = netaddr_ipcache_get(p, name);
+    if (na) {
+      return na;
+    }
+  }
+
   /* Attempt to translate the given name into a pr_netaddr_t using
    * pr_inet_pton() first.
    *
@@ -320,6 +534,8 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
 
     res = pr_inet_pton(AF_INET6, name, &v6.sin6_addr);
     if (res > 0) {
+      int xerrno = errno;
+
       pr_netaddr_set_family(na, AF_INET6);
       pr_netaddr_set_sockaddr(na, (struct sockaddr *) &v6);
       if (addrs)
@@ -327,6 +543,11 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
 
       pr_trace_msg(trace_channel, 7, "'%s' resolved to IPv6 address %s", name,
         pr_netaddr_get_ipstr(na));
+
+      netaddr_ipcache_set(name, na);
+      netaddr_ipcache_set(pr_netaddr_get_ipstr(na), na);
+
+      errno = xerrno;
       return na;
     }
   }
@@ -341,6 +562,8 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
 
   res = pr_inet_pton(AF_INET, name, &v4.sin_addr);
   if (res > 0) {
+    int xerrno = errno;
+
     pr_netaddr_set_family(na, AF_INET);
     pr_netaddr_set_sockaddr(na, (struct sockaddr *) &v4);
     if (addrs)
@@ -348,6 +571,11 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
 
     pr_trace_msg(trace_channel, 7, "'%s' resolved to IPv4 address %s", name,
       pr_netaddr_get_ipstr(na));
+
+    netaddr_ipcache_set(name, na);
+    netaddr_ipcache_set(pr_netaddr_get_ipstr(na), na);
+
+    errno = xerrno;
     return na;
 
   } else if (res == 0) {
@@ -366,11 +594,14 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
 
     hints.ai_family = AF_INET;
     hints.ai_socktype = SOCK_STREAM;
+    hints.ai_protocol = IPPROTO_TCP;
 
     pr_trace_msg(trace_channel, 7,
       "attempting to resolve '%s' to IPv4 address via DNS", name);
     gai_res = pr_getaddrinfo(name, NULL, &hints, &info);
     if (gai_res != 0) {
+      int xerrno = errno;
+
       if (gai_res != EAI_SYSTEM) {
         pr_trace_msg(trace_channel, 1, "IPv4 getaddrinfo '%s' error: %s",
           name, pr_gai_strerror(gai_res));
@@ -378,9 +609,10 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
       } else {
         pr_trace_msg(trace_channel, 1,
           "IPv4 getaddrinfo '%s' system error: [%d] %s", name,
-          errno, strerror(errno));
+          xerrno, strerror(xerrno));
       }
 
+      errno = xerrno;
       return NULL;
     }
 
@@ -393,6 +625,9 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
         info->ai_family == AF_INET ? "IPv4" : "IPv6",
         pr_netaddr_get_ipstr(na));
 
+      netaddr_ipcache_set(name, na);
+      netaddr_ipcache_set(pr_netaddr_get_ipstr(na), na);
+
       pr_freeaddrinfo(info);
     }
 
@@ -409,17 +644,18 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
        * do it this way.)
        */
 
-      gai_res = 0;
-
       memset(&hints, 0, sizeof(hints));
 
       hints.ai_family = AF_INET6;
       hints.ai_socktype = SOCK_STREAM;
+      hints.ai_protocol = IPPROTO_TCP;
 
       pr_trace_msg(trace_channel, 7,
         "attempting to resolve '%s' to IPv6 address via DNS", name);
       gai_res = pr_getaddrinfo(name, NULL, &hints, &info);
       if (gai_res != 0) {
+        int xerrno = errno;
+
         if (gai_res != EAI_SYSTEM) {
           pr_trace_msg(trace_channel, 1, "IPv6 getaddrinfo '%s' error: %s",
             name, pr_gai_strerror(gai_res));
@@ -427,9 +663,10 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
         } else {
           pr_trace_msg(trace_channel, 1, 
             "IPv6 getaddrinfo '%s' system error: [%d] %s", name,
-            errno, strerror(errno));
+            xerrno, strerror(xerrno));
         }
 
+        errno = xerrno;
         return na;
       }
 
@@ -457,6 +694,7 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
 
   pr_trace_msg(trace_channel, 8, "failed to resolve '%s' to an IP address",
     name);
+  errno = ENOENT;
   return NULL;
 }
 
@@ -688,7 +926,7 @@ int pr_netaddr_set_port(pr_netaddr_t *na, unsigned int port) {
   }
 
   errno = EPERM;
-  return 0;
+  return -1;
 }
 
 int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
@@ -721,15 +959,13 @@ int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
     if (pr_netaddr_is_v4mappedv6(na1) == TRUE) {
       tmp_pool = make_sub_pool(permanent_pool);
 
+      pr_trace_msg(trace_channel, 5, "addr '%s' is an IPv4-mapped IPv6 address",
+        pr_netaddr_get_ipstr((pr_netaddr_t *) na1));
+
       /* This case means that na1 is an IPv4-mapped IPv6 address, and
        * na2 is an IPv4 address.
        */
-      a = pr_netaddr_alloc(tmp_pool);
-      pr_netaddr_set_family(a, AF_INET);
-      pr_netaddr_set_port(a, pr_netaddr_get_port(na1));
-      memcpy(&a->na_addr.v4.sin_addr, get_v4inaddr(na1),
-        sizeof(struct in_addr));
-
+      a = pr_netaddr_v6tov4(tmp_pool, na1);
       b = (pr_netaddr_t *) na2;
 
       pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
@@ -739,16 +975,14 @@ int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
     } else if (pr_netaddr_is_v4mappedv6(na2) == TRUE) {
       tmp_pool = make_sub_pool(permanent_pool);
 
+      pr_trace_msg(trace_channel, 5, "addr '%s' is an IPv4-mapped IPv6 address",
+        pr_netaddr_get_ipstr((pr_netaddr_t *) na2));
+
       /* This case means that na is an IPv4 address, and na2 is an
        * IPv4-mapped IPv6 address.
        */
       a = (pr_netaddr_t *) na1;
-
-      b = pr_netaddr_alloc(tmp_pool);
-      pr_netaddr_set_family(b, AF_INET);
-      pr_netaddr_set_port(b, pr_netaddr_get_port(na2));
-      memcpy(&b->na_addr.v4.sin_addr, get_v4inaddr(na2),
-        sizeof(struct in_addr));
+      b = pr_netaddr_v6tov4(tmp_pool, na2);
 
       pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
         "IPv4-mapped IPv6 address '%s'", pr_netaddr_get_ipstr(a),
@@ -768,8 +1002,17 @@ int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
     case AF_INET:
       res = memcmp(&a->na_addr.v4.sin_addr, &b->na_addr.v4.sin_addr,
         sizeof(struct in_addr));
-      if (tmp_pool)
+
+      if (res != 0) {
+        pr_trace_msg(trace_channel, 4, "addr %s does not match addr %s",
+          pr_netaddr_get_ipstr(a), pr_netaddr_get_ipstr(b));
+      }
+
+      if (tmp_pool) {
         destroy_pool(tmp_pool);
+        tmp_pool = NULL;
+      }
+
       return res;
 
 #ifdef PR_USE_IPV6
@@ -777,8 +1020,17 @@ int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
       if (use_ipv6) {
         res = memcmp(&a->na_addr.v6.sin6_addr, &b->na_addr.v6.sin6_addr,
           sizeof(struct in6_addr));
-        if (tmp_pool)
+
+        if (res != 0) {
+          pr_trace_msg(trace_channel, 4, "addr %s does not match addr %s",
+            pr_netaddr_get_ipstr(a), pr_netaddr_get_ipstr(b));
+        }
+
+        if (tmp_pool) {
           destroy_pool(tmp_pool);
+          tmp_pool = NULL;
+        }
+
         return res;
       }
 #endif /* PR_USE_IPV6 */
@@ -826,12 +1078,7 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
       /* This case means that na1 is an IPv4-mapped IPv6 address, and
        * na2 is an IPv4 address.
        */
-      a = pr_netaddr_alloc(tmp_pool);
-      pr_netaddr_set_family(a, AF_INET);
-      pr_netaddr_set_port(a, pr_netaddr_get_port(na1));
-      memcpy(&a->na_addr.v4.sin_addr, get_v4inaddr(na1),
-        sizeof(struct in_addr));
-
+      a = pr_netaddr_v6tov4(tmp_pool, na1);
       b = (pr_netaddr_t *) na2;
 
       pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
@@ -845,12 +1092,7 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
        * IPv4-mapped IPv6 address.
        */
       a = (pr_netaddr_t *) na1;
-
-      b = pr_netaddr_alloc(tmp_pool);
-      pr_netaddr_set_family(b, AF_INET);
-      pr_netaddr_set_port(b, pr_netaddr_get_port(na2));
-      memcpy(&b->na_addr.v4.sin_addr, get_v4inaddr(na2),
-        sizeof(struct in_addr));
+      b = pr_netaddr_v6tov4(tmp_pool, na2);
 
       pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
         "IPv4-mapped IPv6 address '%s'", pr_netaddr_get_ipstr(a),
@@ -988,10 +1230,36 @@ int pr_netaddr_fnmatch(pr_netaddr_t *na, const char *pattern, int flags) {
     const char *ipstr = pr_netaddr_get_ipstr(na);
 
     if (pr_fnmatch(pattern, ipstr, match_flags) == 0) {
-      pr_trace_msg(trace_channel, 6, "DNS name '%s' matches pattern '%s'",
+      pr_trace_msg(trace_channel, 6, "IP address '%s' matches pattern '%s'",
         ipstr, pattern);
       return TRUE;
     }
+
+    /* If the address is an IPv4-mapped IPv6 address, get the IPv4 address
+     * and try to match that against the configured glob pattern.
+     */
+    if (pr_netaddr_is_v4mappedv6(na) == TRUE) {
+      pool *tmp_pool;
+      pr_netaddr_t *a;
+
+      pr_trace_msg(trace_channel, 5, "addr '%s' is an IPv4-mapped IPv6 address",
+        ipstr);
+
+      tmp_pool = make_sub_pool(permanent_pool);
+      a = pr_netaddr_v6tov4(tmp_pool, na);
+
+      ipstr = pr_netaddr_get_ipstr(a);
+
+      if (pr_fnmatch(pattern, ipstr, match_flags) == 0) {
+        pr_trace_msg(trace_channel, 6, "IP address '%s' matches pattern '%s'",
+          ipstr, pattern);
+
+        destroy_pool(tmp_pool);
+        return TRUE;
+      }
+
+      destroy_pool(tmp_pool);
+    }
   }
 
   pr_trace_msg(trace_channel, 4, "addr %s does not match pattern '%s'",
@@ -1051,12 +1319,23 @@ const char *pr_netaddr_get_ipstr(pr_netaddr_t *na) {
 const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
   char *name = NULL;
   char buf[256];
+  pr_netaddr_t *cache = NULL;
 
   if (!na) {
     errno = EINVAL;
     return NULL;
   }
 
+  cache = netaddr_ipcache_get(NULL, pr_netaddr_get_ipstr(na));
+  if (cache &&
+      cache->na_have_dnsstr) {
+    memset(na->na_dnsstr, '\0', sizeof(na->na_dnsstr));
+    sstrncpy(na->na_dnsstr, cache->na_dnsstr, sizeof(na->na_dnsstr));
+    na->na_have_dnsstr = TRUE;
+
+    return na->na_dnsstr;
+  }
+
   /* If this pr_netaddr_t has already been resolved to an DNS string, return the
    * cached string.
    */
@@ -1094,21 +1373,31 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
 #endif /* HAVE_GETHOSTBYNAME2 */
 
       if (hent != NULL) {
-        char **alias;
+        if (hent->h_name != NULL) {
+          netaddr_dnscache_set(pr_netaddr_get_ipstr(na), hent->h_name);
+        }
 
         pr_trace_msg(trace_channel, 10,
           "checking addresses associated with host '%s'",
           hent->h_name ? hent->h_name : "(null)");
-        for (alias = hent->h_aliases; *alias; ++alias) {
-          pr_trace_msg(trace_channel, 10, "host '%s' has alias '%s'",
-            hent->h_name ? hent->h_name : "(null)", *alias);
-        }
-        
+
         switch (hent->h_addrtype) {
           case AF_INET:
             if (family == AF_INET) {
+
               for (checkaddr = hent->h_addr_list; *checkaddr; ++checkaddr) {
                 if (memcmp(*checkaddr, inaddr, hent->h_length) == 0) {
+                  char **alias;
+
+                  for (alias = hent->h_aliases; *alias; ++alias) {
+                    if (hent->h_name) {
+                      pr_trace_msg(trace_channel, 10,
+                        "host '%s' has alias '%s'", hent->h_name, *alias);
+                      netaddr_ipcache_set(*alias, na);
+                      netaddr_dnscache_set(pr_netaddr_get_ipstr(na), *alias);
+                    }
+                  }
+
                   ok = TRUE;
                   break;
                 }
@@ -1119,8 +1408,20 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
 #ifdef PR_USE_IPV6
           case AF_INET6:
             if (use_ipv6 && family == AF_INET6) {
+
               for (checkaddr = hent->h_addr_list; *checkaddr; ++checkaddr) {
                 if (memcmp(*checkaddr, inaddr, hent->h_length) == 0) {
+                  char **alias;
+
+                  for (alias = hent->h_aliases; *alias; ++alias) {
+                    if (hent->h_name) {
+                      pr_trace_msg(trace_channel, 10,
+                        "host '%s' has alias '%s'", hent->h_name, *alias);
+                      netaddr_ipcache_set(*alias, na);
+                      netaddr_dnscache_set(pr_netaddr_get_ipstr(na), *alias);
+                    }
+                  }
+
                   ok = TRUE;
                   break;
                 }
@@ -1143,17 +1444,19 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
             pr_netaddr_get_ipstr(na));
         }
 
-      } else
+      } else {
         pr_log_debug(DEBUG1, "notice: unable to resolve '%s': %s", buf,
           hstrerror(errno));
+      }
     }
 
-  } else
+  } else {
     pr_log_debug(DEBUG10,
       "UseReverseDNS off, returning IP address instead of DNS name");
+  }
 
   if (name) {
-    name = pr_inet_validate(name);
+    name = pr_netaddr_validate_dns_str(name);
 
   } else {
     name = (char *) pr_netaddr_get_ipstr(na);
@@ -1166,15 +1469,54 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
   sstrncpy(na->na_dnsstr, name, sizeof(na->na_dnsstr));
   na->na_have_dnsstr = TRUE;
 
+  /* Update the netaddr object in the cache with the resolved DNS names. */
+  netaddr_ipcache_set(name, na);
+  netaddr_ipcache_set(pr_netaddr_get_ipstr(na), na);
+
   return na->na_dnsstr;
 }
 
+array_header *pr_netaddr_get_dnsstr_list(pool *p, pr_netaddr_t *na) {
+  array_header *res;
+
+  if (p == NULL ||
+      na == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  if (!reverse_dns) {
+    /* If UseReverseDNS is off, then we won't have any names that we trust.
+     * So return an empty list.
+     */
+    return make_array(p, 0, sizeof(char *));
+  }
+
+  res = netaddr_dnscache_get(p, pr_netaddr_get_ipstr(na));
+  if (res == NULL) {
+    res = make_array(p, 0, sizeof(char *));
+  }
+
+  return res;
+}
+
 /* Return the hostname (wrapper for gethostname(2), except returns FQDN). */
 const char *pr_netaddr_get_localaddr_str(pool *p) {
-  char buf[256] = {'\0'};
-  struct hostent *host;
+  char buf[256];
 
+  if (p == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  if (have_localaddr_str) {
+    return pr_netaddr_validate_dns_str(pstrdup(p, localaddr_str));
+  }
+
+  memset(buf, '\0', sizeof(buf));
   if (gethostname(buf, sizeof(buf)-1) != -1) {
+    struct hostent *host;
+
     buf[sizeof(buf)-1] = '\0';
 
     /* Note: this may need to be gethostbyname2() on systems that provide
@@ -1182,20 +1524,30 @@ const char *pr_netaddr_get_localaddr_str(pool *p) {
      * a machine only resolves to an IPv6 address.
      */
     host = gethostbyname(buf);
-
     if (host)
-      return pr_inet_validate(pstrdup(p, host->h_name));
+      return pr_netaddr_validate_dns_str(pstrdup(p, host->h_name));
 
-    return pr_inet_validate(pstrdup(p, buf));
+    return pr_netaddr_validate_dns_str(pstrdup(p, buf));
   }
 
-  pr_trace_msg(trace_channel, 1,  "gethostname(2) error: %s",
-    strerror(errno));
+  pr_trace_msg(trace_channel, 1, "gethostname(2) error: %s", strerror(errno));
   return NULL;
 }
 
+int pr_netaddr_set_localaddr_str(const char *addr_str) {
+  if (addr_str == NULL) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  memset(localaddr_str, '\0', sizeof(localaddr_str));
+  sstrncpy(localaddr_str, addr_str, sizeof(localaddr_str));
+  have_localaddr_str = TRUE;
+  return 0;
+}
+
 int pr_netaddr_is_loopback(const pr_netaddr_t *na) {
-  if (!na) {
+  if (na == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -1207,6 +1559,20 @@ int pr_netaddr_is_loopback(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
+      if (pr_netaddr_is_v4mappedv6(na) == TRUE) {
+        pool *tmp_pool;
+        pr_netaddr_t *v4na;
+        int res;
+
+        tmp_pool = make_sub_pool(permanent_pool);
+        v4na = pr_netaddr_v6tov4(tmp_pool, na);
+
+        res = pr_netaddr_is_loopback(v4na);
+        destroy_pool(tmp_pool);
+
+        return res;
+      }
+
       /* XXX *sigh* Different platforms implement the IN6_IS_ADDR macros
        * differently.  For example, on Linux, those macros expect to operate
        * on s6_addr32, while on Solaris, the macros operate on struct in6_addr.
@@ -1232,15 +1598,15 @@ int pr_netaddr_is_loopback(const pr_netaddr_t *na) {
 /* A slightly naughty function that should go away. It relies too much on
  * knowledge of the internal structures of struct in_addr, struct in6_addr.
  */
-unsigned int pr_netaddr_get_addrno(const pr_netaddr_t *na) {
-  if (!na) {
+uint32_t pr_netaddr_get_addrno(const pr_netaddr_t *na) {
+  if (na == NULL) {
     errno = EINVAL;
-    return -1;
+    return 0;
   }
 
   switch (pr_netaddr_get_family(na)) {
     case AF_INET:
-      return na->na_addr.v4.sin_addr.s_addr;
+      return (uint32_t) na->na_addr.v4.sin_addr.s_addr;
 
 #ifdef PR_USE_IPV6
     case AF_INET6: {
@@ -1254,6 +1620,7 @@ unsigned int pr_netaddr_get_addrno(const pr_netaddr_t *na) {
       int *addrs = ((struct sockaddr_in6 *) pr_netaddr_get_inaddr(na))->s6_addr32;
       return addrs[0];
 #else
+      errno = ENOENT;
       return 0;
 #endif
     }
@@ -1261,7 +1628,7 @@ unsigned int pr_netaddr_get_addrno(const pr_netaddr_t *na) {
   }
 
   errno = EPERM;
-  return -1;
+  return 0;
 }
 
 int pr_netaddr_is_v4mappedv6(const pr_netaddr_t *na) {
@@ -1279,18 +1646,21 @@ int pr_netaddr_is_v4mappedv6(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6: {
+      int res;
+
       if (!use_ipv6) {
         errno = EINVAL;
         return -1;
       }
 
 # ifndef LINUX
-      return IN6_IS_ADDR_V4MAPPED(
+      res = IN6_IS_ADDR_V4MAPPED(
         (struct in6_addr *) pr_netaddr_get_inaddr(na));
 # else
-      return IN6_IS_ADDR_V4MAPPED(
+      res = IN6_IS_ADDR_V4MAPPED(
         ((struct in6_addr *) pr_netaddr_get_inaddr(na))->s6_addr32);
 # endif
+      return res;
     }
 #endif /* PR_USE_IPV6 */
   }
@@ -1299,6 +1669,28 @@ int pr_netaddr_is_v4mappedv6(const pr_netaddr_t *na) {
   return -1;
 }
 
+pr_netaddr_t *pr_netaddr_v6tov4(pool *p, const pr_netaddr_t *na) {
+  pr_netaddr_t *res;
+
+  if (p == NULL ||
+      na == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  if (pr_netaddr_is_v4mappedv6(na) != TRUE) {
+    errno = EPERM;
+    return NULL;
+  }
+
+  res = pr_netaddr_alloc(p);
+  pr_netaddr_set_family(res, AF_INET);
+  pr_netaddr_set_port(res, pr_netaddr_get_port(na));
+  memcpy(&res->na_addr.v4.sin_addr, get_v4inaddr(na), sizeof(struct in_addr));
+
+  return res;
+}
+
 pr_netaddr_t *pr_netaddr_get_sess_local_addr(void) {
   if (have_sess_local_addr) {
     return &sess_local_addr;
@@ -1361,3 +1753,37 @@ void pr_netaddr_enable_ipv6(void) {
   use_ipv6 = 1;
 #endif /* PR_USE_IPV6 */
 }
+
+void pr_netaddr_clear_cache(void) {
+  if (netaddr_iptab) {
+    pr_trace_msg(trace_channel, 5, "emptying netaddr IP cache");
+    (void) pr_table_empty(netaddr_iptab);
+    (void) pr_table_free(netaddr_iptab);
+
+    /* Allocate a fresh table. */
+    netaddr_iptab = pr_table_alloc(netaddr_pool, 0);
+  }
+
+  if (netaddr_dnstab) {
+    pr_trace_msg(trace_channel, 5, "emptying netaddr DNS cache");
+    (void) pr_table_empty(netaddr_dnstab);
+    (void) pr_table_free(netaddr_dnstab);
+
+    /* Allocate a fresh table. */
+    netaddr_dnstab = pr_table_alloc(netaddr_pool, 0);
+  }
+}
+
+void init_netaddr(void) {
+  if (netaddr_pool) {
+    pr_netaddr_clear_cache();
+    destroy_pool(netaddr_pool);
+    netaddr_pool = NULL;
+  }
+
+  netaddr_pool = make_sub_pool(permanent_pool);
+  pr_pool_tag(netaddr_pool, "Netaddr API");
+
+  netaddr_iptab = pr_table_alloc(netaddr_pool, 0);
+  netaddr_dnstab = pr_table_alloc(netaddr_pool, 0);
+}
